@@ -22,7 +22,6 @@ from django.utils import translation
 from .models import (
     Advisor, Album, Band, Bee, Car, Company, Event, Honeycomb, Individual,
     Inventory, Member, MyFileField, Profile, School, Student,
-    UnsafeLimitChoicesTo,
 )
 from .widgetadmin import site as widget_admin_site
 
@@ -182,7 +181,7 @@ class AdminFormfieldForDBFieldTests(SimpleTestCase):
         f = ma.formfield_for_dbfield(Advisor._meta.get_field('companies'), request=None)
         self.assertEqual(
             f.help_text,
-            'Hold down "Control", or "Command" on a Mac, to select more than one.'
+            'Hold down “Control”, or “Command” on a Mac, to select more than one.'
         )
 
 
@@ -587,16 +586,6 @@ class ForeignKeyRawIdWidgetTest(TestCase):
             'Hidden</a></strong>' % {'pk': hidden.pk}
         )
 
-    def test_render_unsafe_limit_choices_to(self):
-        rel = UnsafeLimitChoicesTo._meta.get_field('band').remote_field
-        w = widgets.ForeignKeyRawIdWidget(rel, widget_admin_site)
-        self.assertHTMLEqual(
-            w.render('test', None),
-            '<input type="text" name="test" class="vForeignKeyRawIdAdminField">'
-            '<a href="/admin_widgets/band/?name=%22%26%3E%3Cescapeme&amp;_to_field=id" '
-            'class="related-lookup" id="lookup_id_test" title="Lookup"></a>'
-        )
-
 
 @override_settings(ROOT_URLCONF='admin_widgets.urls')
 class ManyToManyRawIdWidgetTest(TestCase):
@@ -885,7 +874,7 @@ class DateTimePickerSeleniumTests(AdminWidgetSeleniumTestCase):
         for language_code, language_name in settings.LANGUAGES:
             try:
                 catalog = gettext.translation('djangojs', path, [language_code])
-            except IOError:
+            except OSError:
                 continue
             if month_string in catalog._catalog:
                 month_name = catalog._catalog[month_string]
@@ -1404,7 +1393,7 @@ class RelatedFieldWidgetSeleniumTests(AdminWidgetSeleniumTestCase):
 
         # Go ahead and submit the form to make sure it works
         self.selenium.find_element_by_css_selector(save_button_css_selector).click()
-        self.wait_for_text('li.success', 'The profile "changednewuser" was added successfully.')
+        self.wait_for_text('li.success', 'The profile “changednewuser” was added successfully.')
         profiles = Profile.objects.all()
         self.assertEqual(len(profiles), 1)
         self.assertEqual(profiles[0].user.username, username_value)
